/*************** esearch.c ******************
	Contains search and replace commands,
	and marker commands.
	all functions are not system dependent.
*********************************************/

#include <string.h>
#include <ctype.h>

#include "types.h"
#include "sysdep.h"
#include "header.h"
#include "files.h"
#include "esearch.h"
#include "egraphic.h"
#include "menu.h"

#define min(x,y) ((x)<(y)?(x):(y))

char tab[256];
char s_string[64],s_repl[64];
int s_left,s_right,s_cas,s_all,searching;

markertyp marker[MAXMARKER];

char gtoupper (char c)
{
	switch (c)
	{	case '' : return '';
		case '' : return '';
		case '' : return '';
		default : return toupper(c);
	}
}

void make_tab (char *string, char tab[])
/********** make_tab
	make a table of possible jumps for search.
**********/
{	int i,l;
	l=(int)strlen(string);
	for (i=0; i<256; i++) tab[i]=l;
	for (i=0; i<l-1; i++) 
		tab[(unsigned char)string[i]]=
			min(tab[(unsigned char)string[i]],l-i-1);
}

int cstrncmp (char *p1, char *p2, int s)
/******** cstrncmp
	compare two strings, identifying upper and lower case letters.
********/
{	
#ifndef WITTMANN
    while (s-- > 0)
		if (gtoupper(*p1++)!=gtoupper(*p2++)) 
			return (2*(*(p1-1)>*(p2-1))-1);
	return 0;
#else
    p1 += s;
    p2 += s;
    s = -1 - s;
    do
    {
      if( ++s == 0 ) goto found; 	
    } while( gtoupper(p1[s]) != gtoupper(p2[s]));
    return 0;
    found :
    return 1;
#endif
} 

char *search_string (char *start, char *end, 
		char *string, int l, 
		char tab[], int cas)
/******** search_string
	search for the string from start to end (start is the possible
	last position of the string). tab is generated by make_tab. 
	return last position of found string.
	cas is a flag, if upper and lowercase should be different.
********/
{	register char *p=start,last=string[l-1],c;
	int ll=l-1;
	while (1)
	{	if ((c=cas?gtoupper(*p):*p)==last)
			if (cas) 
			{	if (cstrncmp(p-ll,string,l)==0) return p; }
			else 
			{	if (strncmp(p-ll,string,l)==0) return p; }
		p+=tab[(unsigned char)c];
		if (p>=end) return 0;
	}
}

void search (int slot, char *start, char *end, 
	char *string, char tab[], int left, int right, int cas)
/******** search
	Search in that specific file for the string from start to end.
	tab has to be prepared in advance.
	left and right are column limits, for cas see above.
********/
{	int l=(int)strlen(string),x;
	char *p,*line;
	filetyp *ft=files+slot;
	if (l<=0) return;
	marker[MAXMARKER-1].slot=slot;
	marker[MAXMARKER-1].line=end;
	start--;
	do
	{	weiter: start++;
		if (start>=marker[MAXMARKER-1].line) goto endloop;
		if (ft->gap.open && start<ft->gap.start &&
				marker[MAXMARKER-1].line>ft->gap.start)
			end=ft->gap.start;
		else
			end=marker[MAXMARKER-1].line;
		p=search_string(start+l-1,end,string,l,tab,cas);
		if (p==0) 
			if (end==marker[MAXMARKER-1].line) 
			{	refresh_top();
				goto endloop;
			}
			else { start=ft->gap.end-1; goto weiter; }
		start=p-l+1;
		line=prev_line(start+1,ft);
		x=position(line,ft->fend,ft->tabs,start);
	}	while ((x<left-1)||(x>right-1));
	ft->cursor.adr=start;
	ft->cursor.line=line;
	ft->cursor.x=x;
	endloop:
	fredraw(slot);
	marker[MAXMARKER-1].slot=-1;
	edit_line(0); /* bring the line to screen */
}

void replace (int slot, char *start, char *end, 
	char *string, char *repl, char tab[], int left, int right, int cas, 
	int all)
/******** replace
	Search in that specific file for the string from start to end. 
********/
{	int l=(int)strlen(string),x,key,l1;
	char *p,*line;
	filetyp *ft=files+slot;
	if (l<=0) return;
	marker[MAXMARKER-1].slot=slot;
	marker[MAXMARKER-1].line=end;
	start--;
	loop:
	do /* search until found within column limits */
	{	weiter: start++;
		if (start>=marker[MAXMARKER-1].line) goto endloop;
		if (ft->gap.open && start<ft->gap.start && 
				marker[MAXMARKER-1].line>ft->gap.start)
			end=ft->gap.start;
		else
			end=marker[MAXMARKER-1].line;
		p=search_string(start+l-1,end,string,l,tab,cas);
		if (p==0) 
			if (end==marker[MAXMARKER-1].line)
			{	refresh_top();
				goto endloop;
			}
			else { start=ft->gap.end-1; goto weiter; }
		start=p-l+1;
		line=prev_line(start+1,ft);
		x=position(line,ft->fend,ft->tabs,start);
	}	while ((x<left-1)||(x>right-1));
	marker[MAXMARKER-2].line=start;
	marker[MAXMARKER-2].slot=slot;
	update_gap(start,slot);
	if (!all) /* then ask user and update screen */
	{	ft->cursor.adr=marker[MAXMARKER-2].line;
		line=ft->cursor.line=prev_line(ft->cursor.adr+1,ft);
		x=position(line,ft->fend,ft->tabs,marker[MAXMARKER-2].line);
		ft->cursor.x=x;
		fredraw(slot);
		edit_line(0);
		/* show prompt */
		key=form_alert(1,"[2][Replace][Yes|No|Abort]");
		/* read key */
		if (key==1) goto cont;
		else if (key==2) goto loop;
		else goto endloop;
	}
	cont:
	/* do the replacement */
	start=marker[MAXMARKER-2].line;
	l1=(int)strlen(repl);
	if (l1<l) fdelete(slot,start+l1,l-l1);
	else if (l1>l)
		if (!finsert(slot,start+l,l1-l))  goto endloop;
	if (l1) memmove(start,repl,l1);
	ft->flags.updated=1;
	start+=l1-1;
	goto loop;
	endloop:
	/* stop replacement */
	marker[MAXMARKER-1].slot=marker[MAXMARKER-2].slot=-1;
}

void goto_marker (int n) 
/********* goto_marker
	goto marker number n.
*********/
{	int slot;
	filetyp *ft;
	slot=marker[n].slot;
	if ((slot<0)||(!files[slot].flags.used)) return;
	if (slot!=topslot) { topslot=slot; refresh_top(); }
	ft=files+slot;
	ft->cursor.adr=marker[n].line;
	ft->cursor.line=prev_line(marker[n].line+1,ft);
	fredraw(slot);
	edit_line(0);
}

void set_marker (int n)
/******* set_marker
	set the marker number on on top line of active window.
********/
{	int slot;
	slot=get_top();
	if (slot<0) return;
	marker[n].slot=slot;
	marker[n].line=files[slot].cursor.adr;
}
